import sys

import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve
import matplotlib.pyplot as plt

import mesh


class SingularMatrixError(Exception):

    def __init__(self, message):
        self.message = message

def compute_kbar(mesh, K, moyenne):

    Kbar = np.zeros_like(mesh.sommets)
    if moyenne == 'constante':
        Kbar = K * np.ones_like(mesh.sommets)
        Kbar[0] = K
        Kbar[-1] = K
    elif moyenne == 'arithmetique':
        Kbar[1:-1] = (K[:-1] + K[1:])/2
        Kbar[0] = K[0]
        Kbar[-1] = K[-1]
    elif moyenne == 'harmonique':
        Kbar[1:-1] = (mesh.dist[1:-1]*K[:-1]*K[1:]) / ((mesh.sommets[1:-1]-mesh.centres[:-1])*K[1:] + (mesh.centres[1:] - mesh.sommets[1:-1])*K[:-1])
        Kbar[0] = K[0]
        Kbar[-1] = K[-1]
    return Kbar
    
def getmatrhs1D(mesh, K, bclist, rhs, moyenne='harmonique'):
    Kbar = compute_kbar(mesh, K, moyenne)
    tx = Kbar / mesh.dist
    tx[0] = 0
    tx[-1] = 0
    n = len(mesh.centres)
    DiagVecs = np.array([-tx[1:], tx[:-1]+tx[1:], -tx[:-1]])
    A = sp.spdiags(DiagVecs,np.array((-1, 0, 1)), n, n).tocsr()

    if callable(rhs):
        b = mesh.mes*rhs(mesh.centres)
    else:
        b= np.copy(rhs)

# Boundary conditions
    if bclist[0]['type'] == 'Dirichlet':
        A[0,0] += Kbar[0]/mesh.dist[0]
        b[0] += Kbar[0]/mesh.dist[0]*bclist[0]['value']
    elif bclist[0]['type'] == 'Neumann':
        b[0] += bclist[0]['value']
        
    if bclist[-1]['type'] == 'Dirichlet':
        A[-1,-1] += Kbar[-1]/mesh.dist[-1]
        b[-1] += Kbar[-1]/mesh.dist[-1]*bclist[-1]['value']
    elif bclist[-1]['type'] == 'Neumann':
        b[-1] += bclist[-1]['value']

    return A, b


def solve(mesh, K, bclist, rhs, moyenne= 'harmonique'):
    A, b = getmatrhs1D(mesh, K, bclist, rhs, moyenne)
    if bclist[0]['type'] == 'Neumann' and bclist[-1]['type'] == 'Neumann':
        raise SingularMatrixError('Singular matrix')
    u = spsolve(A, b)
    return u

def compute_flux(mesh, K, u, moyenne):
    Kbar = compute_kbar(mesh, K, moyenne)
    tx = Kbar / mesh.dist
    tx[0] = 0
    tx[-1] = 0
    flux = tx[1:-1]*np.diff(u)
    return flux

def test():
    
    L=1
    N=20
    umesh=mesh.make_uniform_mesh(L,N)
    K=3
    bcleftD = {'type':'Dirichlet', 'value':1.}
    bcleftN = {'type':'Neumann', 'value':-6.}
    bcrightD = {'type':'Dirichlet', 'value':2.}
    bcrightN = {'type':'Neumann', 'value':0.}
    bclist = [bcleftD, bcrightD]
    
    rhsfun = lambda x: 6 * np.ones_like(x)
    solex = lambda x: x * (1 - x) + x + 1
    
    u = solve(umesh, K, bclist, rhsfun, 'constante')
    
    x = umesh.centres
    uex = solex(x)
    erreur = np.linalg.norm(u-uex,2) / np.linalg.norm(uex,2)
    print("erreur= ", erreur)

    
if __name__ == '__main__':
    test()
